﻿/** 
 * @file MainForm.cs
 * @author Yarrago
 * @note Copyright © Yarrago (yarrago.com) 2010
 * @version 1.0.3
 * @date Created: 31/05/2010
 * @date Modified: 30/11/2010
 * @note Status: Complete
 * 
 * @note Modified by Yarrago on 04/06/2010. Description: Completed main code.
 * @note Modified by Yarrago on 07/06/2010. Description: Added prepend date option.
 * @note Modified by Yarrago on 25/11/2010. Description: Modified date fromat from yyyy_MM_dd to yyyy-MM-dd. Added sub folder option.
 * @note Modified by Yarrago on 30/11/2010. Description: Allowed user to actually delete files. Added an option to allow confirmation of completion.
 * 
 * 
 * 
 * 
 * 
 * License: "Modified BSD License" 3-clause license.
 * 
 * Copyright (c) 2010, Yarrago (yarrago.com)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *   - Redistributions of source code must retain the above copyright notice, this list of 
 *     conditions and the following disclaimer.
 *   - Redistributions in binary form must reproduce the above copyright notice, this list of 
 *     conditions and the following disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   - Neither the name of the yarrago.com nor the names of its contributors may be used to endorse
 *     or promote products derived from this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

// References Required:
//      ReachFramework
//      PresentationCore
//      PresentationFramework
//      WindowsBase

namespace PrintToImageConverter
{
    public partial class MainForm : Form
    {
        /// <summary>
        /// Represents the image format.
        /// </summary>
        enum ImageType : uint
        {
            bmp,
            png,
            jpg
        }

        /// <summary>
        /// Represents the options for post processing.
        /// </summary>
        enum PostProcessType : uint
        {
            Nothing = 0,
            Single = 1,
            AllSequential = 2,
            AllSimultaneous = 3
        }

        private const string bmpExt = ".bmp";                                   // The extension to associate with bitmap files.
        private const string pngExt = ".png";                                   // The extension to associate with png files.
        private const string jpgExt = ".jpg";                                   // The extension to associate with jpg fiels.

        private const string imageFilterStart = "Image Files";                  // 
        private const string exeFilter = "Executable Files (*.exe)|*.exe";      //
        private const string xpsFilter = "XPS Files (*.xps)|*.xps";             // 

        private const int AutoStartTime = 5;                                    // The number of seconds to start auto processing after the application starts or auto processing is enabled.

        System.Windows.Controls.Image previewImage = new System.Windows.Controls.Image();   // The control to hold the preview image (couldn't add from the designer as it sits in a migration container on the form.

        // All of the following variables represent the current application state and sould only be used directly from the associated properties.
        // The inital constant are never used as they are all loaded from the settings file start up first time and everytime.

        private string inputFilename_ = "";                                     // The input (.xps) filename.

        private string alternateApplication_ = "";                              // The full filename and path of the alternate application to start to handel the xps file.

        private uint currentPage_ = 0;                                          // The page number for the current page displayed in the image preview.

        private ImageType outputImageFileType_ = ImageType.bmp;                 // The type of file we are converting the xps file to.
        private uint outputDPI_ = 300;                                          // The DPI of the output image.
        private System.Windows.Media.Imaging.PngInterlaceOption pngInterlace_ = System.Windows.Media.Imaging.PngInterlaceOption.Default;    // Whether to interlace the output png or not.
        private uint jpgQuality_ = 100;                                         // The JPG quality setting for the output JPG.
        private bool useInputFilename_ = false;                                 // Whether to use the input filename to generate the output filename instead of OutputFilenameBase.
        private bool createSubFolder_ = false;                                  // Whether to create a sub folder when using the input filename to generate the ouput filename.
        private bool useCurrentDateInFilename_ = false;                         // Whether to use current date when generating the filename.
        private string outputFilenameBase_ = "";                                // The output filenames base (the filename for the output file without a unique identifier added).

        private PostProcessType postProcess_ = PostProcessType.Nothing;         // What option has been selected for processing the output file after it has been created.
        private string postProcessApplication_ = "";                            // The application to use when post processing.

        private bool delayProcess_ = true;                                      // Whether to delay the post processing between runs when using AllSequential option.
        private float delaySeconds_ = 5;                                        // The number of seconds to delay when between runs when AllSequential is selected and DelayProcess is enabled. 
        private bool deleteWhenDone_ = false;                                   // Whether to delete the input file after it has been processed.
        private bool autoStart_ = true;                                         // Whether to auto start processing.
        private bool autoClose_ = true;                                         // Whether to close the program after the input file has been successfully processed.
        private bool confirmProcessed_ = true;                                  // Whether to prompt the user when the conversion process has completed.


        // ========================================================================================
        // Access And Update Code
        // ========================================================================================
        
        /// <summary>
        /// Gets or sets the input filename.
        /// </summary>
        private string InputFilename
        {
            get
            {
                return inputFilename_;
            }
            set
            {
                inputFilename_ = value;
                inputFilenameLabel.Text = inputFilename_;
            }
        }

        /// <summary>
        /// Gets or sets the full filename and path of the alternate application to start to handel the xps file.
        /// </summary>
        private string AlternateApplication
        {
            get
            {
                return alternateApplication_;
            }
            set
            {
                alternateApplication_ = value;
                alternateApplicationTextBox.Text = alternateApplication_;
            }
        }

        /// <summary>
        /// Gets or sets the page number for the current page displayed in the image preview.
        /// </summary>
        private uint CurrentPage
        {
            get
            {
                return currentPage_;
            }
            set
            {
                uint pages = getXPSPages(InputFilename);

                if (pages == 0)
                {
                    currentPage_ = 0;
                }
                else
                {
                    currentPage_ = value % pages;
                }

                // TODO Consider implementing this functionality. Not important, probably not used.
                // TODO Update current page display.
                MessageBox.Show("Not Implemented.");
            }
        }

        /// <summary>
        /// Gets or sets the type of file we are converting the xps file to.
        /// </summary>
        private ImageType OutputImageFileType
        {
            get
            {
                return outputImageFileType_;
            }
            set
            {
                bool showPNGOptions = false;
                bool showJPGOptions = false;
                bool valueChanged = (outputImageFileType_ != value);

                outputImageFileType_ = value;

                switch (outputImageFileType_)
                {
                    case ImageType.bmp:
                        outputBMPRadioButton.Checked = true;
                        break;
                    case ImageType.png:
                        outputPNGRadioButton.Checked = true;
                        showPNGOptions = true;
                        break;
                    case ImageType.jpg:
                        outputJPGRadioButton.Checked = true;
                        showJPGOptions = true;
                        break;
                }

                interlaceCheckBox.Visible = showPNGOptions;
                imageQualityLabel.Visible = showJPGOptions;
                imageQualityNumericUpDown.Visible = showJPGOptions;

                if (valueChanged)
                {
                    updateOutputFilenameExtension();
                }
            }
        }

        /// <summary>
        /// Gets or sets the DPI of the output image.
        /// </summary>
        private uint OutputDPI
        {
            get
            {
                return outputDPI_;
            }
            set
            {
                outputDPI_ = value;
                dpiNumericUpDown.Value = outputDPI_;
            }
        }

        /// <summary>
        /// Gets or sets whether to interlace the output png or not.
        /// </summary>
        private System.Windows.Media.Imaging.PngInterlaceOption PngInterlace
        {
            get
            {
                return pngInterlace_;
            }
            set
            {
                pngInterlace_ = value;

                switch (pngInterlace_) //)
                {
                    case System.Windows.Media.Imaging.PngInterlaceOption.On:
                        interlaceCheckBox.CheckState = CheckState.Checked;
                        break;
                    case System.Windows.Media.Imaging.PngInterlaceOption.Off:
                        interlaceCheckBox.CheckState = CheckState.Unchecked;
                        break;
                    case System.Windows.Media.Imaging.PngInterlaceOption.Default:
                        interlaceCheckBox.CheckState = CheckState.Indeterminate;
                        break;
                }
            }
        }

        /// <summary>
        /// Gets or sets the JPG quality setting for the output JPG.
        /// </summary>
        private uint JpgQuality
        {
            get
            {
                return jpgQuality_;
            }
            set
            {
                jpgQuality_ = value;

                imageQualityNumericUpDown.Value = value;
            }
        }

        /// <summary>
        /// Gets or sets whether to use the input filename to generate the output filename instead of OutputFilenameBase.
        /// </summary>
        private bool UseInputFilename
        {
            get
            {
                return useInputFilename_;
            }
            set
            {
                useInputFilename_ = value;
                useInputFilenameCheckBox.Checked = useInputFilename_;
                outputFilenameButton.Enabled = !useInputFilename_;
                outputFilenameTextBox.Enabled = !useInputFilename_;
                createSubFolderCheckBox.Enabled = useInputFilename_;
            }
        }

        /// <summary>
        /// Gets or sets whether to create a sub folder when generating input filename.
        /// </summary>
        private bool CreateSubFolder
        {
            get
            {
                return createSubFolder_;
            }
            set
            {
                createSubFolder_ = value;
                createSubFolderCheckBox.Checked = createSubFolder_;
            }
        }

        /// <summary>
        /// Whether to use current date when generating the filename.
        /// </summary>
        private bool UseCurrentDateInFilename
        {
            get
            {
                return useCurrentDateInFilename_;
            }
            set
            {
                useCurrentDateInFilename_ = value;

                currentDateCheckBox.Checked = useCurrentDateInFilename_;
            }
        }

        /// <summary>
        /// Gets or sets the output filenames base (the filename for the output file without a unique identifier added).
        /// </summary>
        private string OutputFilenameBase
        {
            get
            {
                return outputFilenameBase_;
            }
            set
            {
                outputFilenameBase_ = value;
                outputFilenameTextBox.Text = outputFilenameBase_;
            }
        }

        /// <summary>
        /// Gets or sets what option has been selected for processing the output file after it has been created.
        /// </summary>
        private PostProcessType PostProcess
        {
            get
            {
                return postProcess_;
            }
            set
            {
                bool enableDelay = false;
                postProcess_ = value;

                switch (postProcess_)
                {
                    case PostProcessType.Nothing:
                        postProcessNothingRadioButton.Checked = true;
                        enableDelay = false;
                        break;
                    case PostProcessType.Single:
                        postProcessSingleRadioButton.Checked = true;
                        enableDelay = false;
                        break;
                    case PostProcessType.AllSequential:
                        postProcessAllSequentialRadioButton.Checked = true;
                        enableDelay = true;
                        break;
                    case PostProcessType.AllSimultaneous:
                        postProcessAllSimultaneousRadioButton.Checked = true;
                        enableDelay = false;
                        break;
                }

                delayProcessCheckBox.Enabled = enableDelay;
                delayNumericUpDown.Enabled = enableDelay & delayProcessCheckBox.Checked;
            }
        }

        /// <summary>
        /// Gets or sets the application to use when post processing.
        /// </summary>
        private string PostProcessApplication
        {
            get
            {
                return postProcessApplication_;
            }
            set
            {
                postProcessApplication_ = value;
                postProcessTextBox.Text = postProcessApplication_;
            }
        }

        /// <summary>
        /// Gets or sets whether to delay the post processing between runs when using AllSequential option.
        /// </summary>
        private bool DelayProcess
        {
            get
            {
                return delayProcess_;
            }
            set
            {
                delayProcess_ = value;
                delayProcessCheckBox.Checked = delayProcess_;
                delayNumericUpDown.Enabled = delayProcess_ & (PostProcess == PostProcessType.AllSequential);
            }
        }

        /// <summary>
        /// Gets or sets the number of seconds to delay when between runs when AllSequential is selected and DelayProcess is enabled. 
        /// </summary>
        private float DelaySeconds
        {
            get
            {
                return delaySeconds_;
            }
            set
            {
                delaySeconds_ = value;
                delayNumericUpDown.Value = (decimal)delaySeconds_;
            }
        }

        /// <summary>
        /// Gets or sets whether to delete the input file after it has been processed.
        /// </summary>
        private bool DeleteWhenDone
        {
            get
            {
                return deleteWhenDone_;
            }
            set
            {
                deleteWhenDone_ = value;
                deleteInputCheckBox.Checked = deleteWhenDone_;
            }
        }

        /// <summary>
        /// Gets or sets whether to auto start processing.
        /// </summary>
        private bool AutoStart
        {
            get
            {
                return autoStart_;
            }
            set
            {
                autoStart_ = value;
                autoStartCheckBox.Checked = autoStart_;

                autoStartTimer.Interval = AutoStartTime * 1000;
                autoStartTimer.Enabled = autoStart_;
            }
        }

        /// <summary>
        /// Gets or sets whether to close the program after the input file has been successfully processed.
        /// </summary>
        private bool AutoClose
        {
            get
            {
                return autoClose_;
            }
            set
            {
                autoClose_ = value;
                autoCloseCheckBox.Checked = autoClose_;
            }
        }

        /// <summary>
        /// Gets or sets whether to confirm the file has been processed after the input file has been successfully processed.
        /// </summary>
        private bool ConfirmProcessed
        {
            get
            {
                return confirmProcessed_;
            }
            set
            {
                confirmProcessed_ = value;
                confimProcessedCheckBox.Checked = confirmProcessed_;
            }
        }

        // ========================================================================================
        // Interface Code
        // ========================================================================================

        /// <summary>
        /// MainForm constructor used to initalise the display elements.
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            setUpPreview();
        }

        /// <summary>
        /// Called when the main form is loaded.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_Load(object sender, EventArgs e)
        {
            string[] cmd = Environment.GetCommandLineArgs();    // First argument will be the executable path/name and the seconds can be a filename to load.
            if (cmd.Length >= 2)
            {
                if (System.IO.File.Exists(cmd[1]))
                {
                    InputFilename = cmd[1];
                }
            }

            loadSettings();

            // If a file has been loaded
            if (System.IO.File.Exists(InputFilename))
            {
                // Default the selected button to process.
                instantProcessButton.Select();
            }
            else
            {
                // Otherwise default to load file button.
                loadFileButton.Select();
            }
        }

        /// <summary>
        /// Called when the main form is closed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            saveSettings();
        }

        /// <summary>
        /// Called when the load file button is clicked. Is used to load a new xps file into the program.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void loadFileButton_Click(object sender, EventArgs e)
        {
            openFileDialog.FileName = InputFilename;
            openFileDialog.Filter = xpsFilter;

            DialogResult r = openFileDialog.ShowDialog();

            if (r == DialogResult.OK)
            {
                InputFilename = openFileDialog.FileName;
            }
        }

        /// <summary>
        /// Called when the text in the alternateApplicationTextBox is changed. Is used to update the internal application state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void alternateApplicationTextBox_TextChanged(object sender, EventArgs e)
        {
            AlternateApplication = alternateApplicationTextBox.Text;
        }

        /// <summary>
        /// Called when the alternateApplicationPickerButton is clicked. It is used to allow the user to select the application to be used.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void alternateApplicationPickerButton_Click(object sender, EventArgs e)
        {
            openFileDialog.FileName = alternateApplicationTextBox.Text;
            openFileDialog.Filter = exeFilter;

            DialogResult r = openFileDialog.ShowDialog();

            if (r == DialogResult.OK)
            {
                AlternateApplication = openFileDialog.FileName;
            }
        }

        /// <summary>
        /// Called when the alternateApplicationRunButton is clicked. It is used to allow the user to load the xps file into another xps enabled application and stops processing in this application.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void alternateApplicationRunButton_Click(object sender, EventArgs e)
        {
            bool success = false;

            if (System.IO.File.Exists(InputFilename))
            {
                success = startProcess(AlternateApplication, wrapFilename(InputFilename));  // Wrap the filename in quotes incase it contains spaces.
            }

            if (success)
            {
                Application.Exit();
            }
        }

        /// <summary>
        /// Called when the previousPageButton is clicked. It is used to change the preview image to the previous page.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void previousPageButton_Click(object sender, EventArgs e)
        {
            CurrentPage--;
        }

        /// <summary>
        /// Called when the nextPageButton is clicked. It is used to change the preview image to the next page.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void nextPageButton_Click(object sender, EventArgs e)
        {
            CurrentPage++;
        }

        /// <summary>
        /// Called when the outputBMPRadioButton state changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void outputBMPRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            updateOutputImageFileType();
        }

        /// <summary>
        /// Called when the outputPNGRadioButton state changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void outputPNGRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            updateOutputImageFileType();
        }

        /// <summary>
        /// Called when the outputJPGRadioButton state changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void outputJPGRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            updateOutputImageFileType();
        }

        /// <summary>
        /// Called when the dpiNumericUpDown value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void dpiNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            OutputDPI = (uint)dpiNumericUpDown.Value;
        }

        /// <summary>
        /// Called when the interlaceCheckBox value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void interlaceCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            switch (interlaceCheckBox.CheckState)
            {
                case CheckState.Checked:
                    PngInterlace = System.Windows.Media.Imaging.PngInterlaceOption.On;
                    break;
                case CheckState.Unchecked:
                    PngInterlace = System.Windows.Media.Imaging.PngInterlaceOption.Off;
                    break;
                case CheckState.Indeterminate:
                    PngInterlace = System.Windows.Media.Imaging.PngInterlaceOption.Default;
                    break;
            }
        }

        /// <summary>
        /// Called when the imageQualityNumericUpDown value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void imageQualityNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            JpgQuality = (uint)imageQualityNumericUpDown.Value;
        }

        /// <summary>
        /// Called when the currentDateCheckBox value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void currentDateCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            UseCurrentDateInFilename = currentDateCheckBox.Checked;
        }

        /// <summary>
        /// Called when the useInputFilenameCheckBox value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void useInputFilenameCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            UseInputFilename = useInputFilenameCheckBox.Checked;
        }

        /// <summary>
        /// Called when the createSubFolderCheckBox value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void createSubFolderCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            CreateSubFolder = createSubFolderCheckBox.Checked;
        }

        /// <summary>
        /// Called when the outputFilenameTextBox value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void outputFilenameTextBox_TextChanged(object sender, EventArgs e)
        {
            OutputFilenameBase = outputFilenameTextBox.Text;
        }

        /// <summary>
        /// Called when the outputFileButtonButton is clicked. It is used to allow the user to select the output filename.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void outputFileButton_Click(object sender, EventArgs e)
        {
            string ext = getImageExt();
            saveFileDialog.FileName = outputFilenameTextBox.Text;
            saveFileDialog.Filter = imageFilterStart + @" (*" + ext + @") |*" + ext;

            DialogResult r = saveFileDialog.ShowDialog();

            if (r == DialogResult.OK)
            {
                OutputFilenameBase = saveFileDialog.FileName;
            }
        }

        /// <summary>
        /// Called when the postProcessNothingRadioButton value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void postProcessNothingRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            updatePostProcess();
        }

        /// <summary>
        /// Called when the postProcessSingleRadioButton value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void postProcessSingleRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            updatePostProcess();
        }

        /// <summary>
        /// Called when the postProcessAllSequentialRadioButton value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void postProcessAllSequentialRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            updatePostProcess();
        }

        /// <summary>
        /// Called when the postProcessAllSequentialRadioButton value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void postProcessAllSequentialRadioButton_Click(object sender, EventArgs e)
        {
            if (postProcessAllSequentialRadioButton.Checked)
            {
                MessageBox.Show("Warning, this can cause many instances of a program to load which may make your computer unresponsive. Please only use this option if you know what you are doing.");
            }
        }

        /// <summary>
        /// Called when the postProcessAllSimultaneousRadioButton value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void postProcessAllSimultaneousRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            updatePostProcess();
        }

        /// <summary>
        /// Called when the postProcessTextBox text changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void postProcessTextBox_TextChanged(object sender, EventArgs e)
        {
            PostProcessApplication = postProcessTextBox.Text;
        }

        /// <summary>
        /// Called when the postProcessPickerButton is clicked. Is used to allow the user to select an application to run after the xps document has been converted to an image.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void postProcessPickerButton_Click(object sender, EventArgs e)
        {
            openFileDialog.FileName = postProcessTextBox.Text;
            openFileDialog.Filter = exeFilter;

            DialogResult r = openFileDialog.ShowDialog();

            if (r == DialogResult.OK)
            {
                PostProcessApplication = openFileDialog.FileName;
            }

        }

        /// <summary>
        /// Called when the postProcessRunButton is clicked. Is used to allow the user to load the external application if they want to quickly load it for any reason.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void postProcessRunButton_Click(object sender, EventArgs e)
        {
            startProcess(PostProcessApplication, "");
        }

        /// <summary>
        /// Called when the delayProcessCheckBox value changes. Is used to update the internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void delayProcessCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            DelayProcess = delayProcessCheckBox.Checked;
        }

        /// <summary>
        /// Called when the delayNumericUpDown value changes. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void delayNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            DelaySeconds = (float)delayNumericUpDown.Value;
        }

        /// <summary>
        /// Called when the autoStartCheckBox changes state. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void autoStartCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            AutoStart = autoStartCheckBox.Checked;
        }

        /// <summary>
        /// Called when the autoCloseCheckBox changes state. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void autoCloseCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            AutoClose = autoCloseCheckBox.Checked;
        }

        /// <summary>
        /// Called when the deleteInputCheckBox changes state. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void deleteInputCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            DeleteWhenDone = deleteInputCheckBox.Checked;
        }

        /// <summary>
        /// Called when the confimProcessedCheckBox changes state. Is used to update the applications internal state.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void confimProcessedCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            ConfirmProcessed = confimProcessedCheckBox.Checked;
        }

        /// <summary>
        /// Called when the autoStartTimer timer expires. Used to start processing the xps input file.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void autoStartTimer_Tick(object sender, EventArgs e)
        {
            process();
        }

        /// <summary>
        /// Called when the instantProcessButton is clicked. Is used by the user to manually start the processing of the xps input file.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void instantProcessButton_Click(object sender, EventArgs e)
        {
            process();
        }


        // ========================================================================================
        // Functionality Code
        // ========================================================================================

        /// <summary>
        /// Starts processing of the xps document.
        /// 
        /// This function is designed to be the top part of the process and hadel the form state.
        /// </summary>
        private void process()
        {
            autoStartTimer.Enabled = false;
            fixRootOutputFilenameBase();

            convertXPSToImage();
        }

        /// <summary>
        /// Converts the XPS document using the current state.
        /// </summary>
        private void convertXPSToImage()
        {
            // This should really run in its own thread.
            // http://msdn.microsoft.com/en-us/library/ts553s52.aspx
            // Not as easy as it sounds, WPF does not interact well in threads.
            // See: "The calling thread must be STA, because many UI components require this."

            bool success = false;
            uint pages = 0;

            string inputFilename = InputFilename;
            string nextFileName = "";
            List<string> filenames = new List<string>();

            if (System.IO.File.Exists(inputFilename))
            {
                try
                {
                    System.Windows.Xps.Packaging.XpsDocument xpsDoc = new System.Windows.Xps.Packaging.XpsDocument(inputFilename, System.IO.FileAccess.Read);
                    System.Windows.Documents.FixedDocumentSequence fds = xpsDoc.GetFixedDocumentSequence();

                    updateProgress(0, pages);

                    // For all pages.
                    pages = (uint)fds.DocumentPaginator.PageCount;
                    for (uint i = 0; i < pages; i++)
                    {
                        // Get a page.
                        System.Windows.Documents.DocumentPage page = fds.DocumentPaginator.GetPage((int)i);

                        const double diu = 96.0; // Device-Independent Units - A WPF Constant

                        // Create a render target.
                        System.Windows.Media.Imaging.RenderTargetBitmap renderTarget = new System.Windows.Media.Imaging.RenderTargetBitmap((int)(page.Size.Width*(OutputDPI/diu)), (int)(page.Size.Height*(OutputDPI/diu)),
                                                                                                                                           OutputDPI, OutputDPI,
                                                                                                                                           System.Windows.Media.PixelFormats.Default);
                        // Render the page (to a memory bitmap I assume).
                        renderTarget.Render(page.Visual);
                        
                        // Display a preview of the image.
                        loadPreview(renderTarget);

                        // Generate an encoder (to format the memory bitmip into the encoding that we will save).
                        System.Windows.Media.Imaging.BitmapEncoder encoder = generateImageFileEncoder();

                        // Add the image to the encoder (clear the frames just to be sure, should not need to).
                        encoder.Frames.Clear();
                        encoder.Frames.Add(System.Windows.Media.Imaging.BitmapFrame.Create(renderTarget));

                        // Generate the next file name.
                        nextFileName = generateImageFileName();
                        if (nextFileName == "")
                        {
                            MessageBox.Show("Unable to generate filename to save image.");
                            break;
                        }
                        else
                        {
                            // Warn the user if the file already exists and give them a chance to abort.
                            // Should not happen, but is a last chance safeguard.
                            if (System.IO.File.Exists(nextFileName))
                            {
                                DialogResult r = MessageBox.Show("The converted filename " + nextFileName + " already exists, overwrite?", "Confirm Overwrite", MessageBoxButtons.YesNo);

                                if (r != DialogResult.Yes)
                                {
                                    break;
                                }
                            }

                            // Create the output directory if it doesn't exist.
                            if (!System.IO.Directory.Exists(nextFileName))
                            {
                                System.IO.Directory.CreateDirectory(System.IO.Path.GetDirectoryName(nextFileName));
                            }


                            System.IO.FileStream fileStream = new System.IO.FileStream(nextFileName, System.IO.FileMode.Create, System.IO.FileAccess.Write);

                            // Save the image and close the file.
                            encoder.Save(fileStream);
                            fileStream.Close();

                            // Add the name to list of files that we have created (for post processing if needed).
                            filenames.Add(nextFileName);
                        }

                        updateProgress(i+1, pages, nextFileName);
                        System.Threading.Thread.Sleep(0);
                    }

                    xpsDoc.Close();

                    success = true;
                }
                catch (Exception e)
                {
                    MessageBox.Show("Error converting file.\n\n Error: " + e.Message);
                    // Don't know don't care. No idea what the error was, just warn the user for now.
                }
            }

            if (success)
            {
                postProcess(filenames);

                if (DeleteWhenDone)
                {
                    System.IO.FileInfo fi = new System.IO.FileInfo(inputFilename);
                    bool deleted = ShellInteraction.FileFolderShellInteraction.recycle(fi, this.Handle, true);
                    if (!deleted || System.IO.File.Exists(inputFilename))
                    {
                        MessageBox.Show("Unable to delete input file. \n\n" + inputFilename);
                        success = false;
                    }
                }

                if (ConfirmProcessed)
                {
                    MessageBox.Show("Processed!");
                }
            }

            if ((success) && (AutoClose))
            {
                Application.Exit();
            }
        }

        /// <summary>
        /// Handels post processing of the output images by an external application.
        /// </summary>
        /// <param name="filenames">The list of filenames to process in the external application.</param>
        /// <returns></returns>
        private bool postProcess(List<string> filenames)
        {
            bool success = false;
            string argument = "";

            switch (PostProcess)
            {
                case PostProcessType.Nothing:
                    // Do nothing...That was easy.
                    success = true;
                    break;
                case PostProcessType.Single:
                    // Load the first file.
                    // Make sure there is one.
                    if (filenames.Count >= 1)
                    {
                        // Get it.
                        argument = filenames[0];

                        // Check it exists.
                        if (System.IO.File.Exists(argument))
                        {
                            success = startProcess(PostProcessApplication, wrapFilename(argument)); // Wrap the filename with quotes incase it includes spaces.
                        }
                    }
                    break;
                case PostProcessType.AllSequential:
                    // Load each file sequentially.
                    for (int i = 0; i < filenames.Count; i++)
                    {
                        if (i == 0)
                        {
                            success = true;
                        }

                        argument = filenames[i];

                        if (System.IO.File.Exists(argument))
                        {
                            success &= startProcess(PostProcessApplication, wrapFilename(argument)); // Wrap the filename with quotes incase it includes spaces.
                        }

                        if (DelayProcess)
                        {
                            System.Threading.Thread.Sleep((int)(DelaySeconds * 1000));
                        }
                    }
                    break;
                case PostProcessType.AllSimultaneous:
                    for (int i = 0; i < filenames.Count; i++)
                    {
                        if (System.IO.File.Exists(filenames[i]))
                        {
                            argument += wrapFilename(filenames[i]) + " ";  // Wrap the filename with quotes incase it includes spaces.
                        }
                    }

                    success = startProcess(PostProcessApplication, argument);
                    break;
            }

            return success;
        }

        /// <summary>
        /// Creates an image file encoder for the desired type of output image.
        /// </summary>
        /// <returns>An encoder for the type of image to be generated.</returns>
        private System.Windows.Media.Imaging.BitmapEncoder generateImageFileEncoder()
        {
            System.Windows.Media.Imaging.BitmapEncoder encoder;
            ImageType imageType = getImageFileType();

            // Create the encoder based on the desiered file type.
            switch (imageType)
            {
                case ImageType.bmp:
                default:
                    encoder = new System.Windows.Media.Imaging.BmpBitmapEncoder();
                    break;
                case ImageType.png:
                    System.Windows.Media.Imaging.PngBitmapEncoder penc = new System.Windows.Media.Imaging.PngBitmapEncoder();
                    penc.Interlace = PngInterlace;
                    encoder = penc;
                    break;
                case ImageType.jpg:
                    System.Windows.Media.Imaging.JpegBitmapEncoder jenc = new System.Windows.Media.Imaging.JpegBitmapEncoder();
                    jenc.QualityLevel = (int)JpgQuality;
                    encoder = jenc;
                    break;
            }

            return encoder;
        }

        /// <summary>
        /// Generates a new filename that is not used based filename base. 
        /// </summary>
        /// <returns>The new unused filename ("" on error).</returns>
        private string generateImageFileName()
        {
            bool valid = false;
            string filenameBase = OutputFilenameBase;
            string filename = "";

            if (UseInputFilename)
            {
                filenameBase = System.IO.Path.ChangeExtension(InputFilename, getImageExt());
            }
            
            if (UseCurrentDateInFilename)
            {
                string f = DateTime.Now.ToString("yyyy-MM-dd ") + System.IO.Path.GetFileName(filenameBase);
                filenameBase = System.IO.Path.Combine(System.IO.Path.GetDirectoryName(filenameBase), f);
            }

            if (UseInputFilename)
            {
                if (CreateSubFolder)
                {
                    string subfolder = System.IO.Path.Combine(System.IO.Path.GetDirectoryName(filenameBase), System.IO.Path.ChangeExtension(filenameBase, null));
                    filenameBase = System.IO.Path.Combine(subfolder, System.IO.Path.GetFileName(filenameBase));
                }
            }

            string extension = System.IO.Path.GetExtension(filenameBase);
            filenameBase = System.IO.Path.ChangeExtension(filenameBase, null);
            for (uint i = 1; i < uint.MaxValue; i++)
            {
                filename = filenameBase + uintToSequenceString(i) + extension;

                if (!System.IO.File.Exists(filename))
                {
                    valid = true;
                    break;
                }
            }

            if (!valid)
            {
                filename = "";
            }

            return filename;
        }

        /// <summary>
        /// Converts the number to a string sequence number of the format "+0000". That is a 4 digit sequence number padded with zeros or if number is bigger then 4 digits all the digits.
        /// 
        /// EG input = 0,       output =  0000
        /// EG input = 1,       output =  0001
        /// EG input = 101,     output =  0101
        /// EG input = 10101,   output = 10101
        /// </summary>
        /// <param name="number">The number to be converted.</param>
        /// <returns>The converted string sequence number.</returns>
        static private string uintToSequenceString(uint number)
        {
            return " " + number.ToString().PadLeft(4, '0');
        }

        /// <summary>
        /// Gets the number of pages in the xps document.
        /// </summary>
        /// <param name="inputFilename">The filename of the document to get the number of pages from.</param>
        /// <returns>The number of pages in the xps document.</returns>
        static private uint getXPSPages(string inputFilename)
        {
            uint pageCount = 0;

            if (System.IO.File.Exists(inputFilename))
            {
                try
                {
                    System.Windows.Xps.Packaging.XpsDocument xpsDoc = new System.Windows.Xps.Packaging.XpsDocument(inputFilename, System.IO.FileAccess.Read);
                    System.Windows.Documents.FixedDocumentSequence fds = xpsDoc.GetFixedDocumentSequence();
                    pageCount = (uint)fds.DocumentPaginator.PageCount;
                }
                catch (Exception)
                {
                    // Don't know don't care. No idea what the error was, just do nothing for now.
                }
            }

            return pageCount;
        }

        /// <summary>
        /// Wraps the filename in "". For use with startProcess() to ensure the filename is not split into seperate arguments if it contains spaces.
        /// </summary>
        /// <param name="filename">The filename to be used.</param>
        /// <returns>The wrapped filename</returns>
        private static string wrapFilename(string filename)
        {
            return "\"" + filename + "\"";
        }

        /// <summary>
        /// Starts a new process. Used to load other applications with command line arguments.
        /// </summary>
        /// <param name="filename">The filename of the other application.</param>
        /// <param name="argument">The argument(s) as a string of space separated arguments. Use wrapFilename to wrap arguments that may contain spaces to ensure they arn't split into seperate arguments.</param>
        /// <returns>Whether the process was able to be started successfully.</returns>
        private static bool startProcess(string filename, string argument)
        {
            bool success = false;

            if (System.IO.File.Exists(filename))
            {
                try
                {
                    System.Diagnostics.Process p = System.Diagnostics.Process.Start(filename, argument);
                    success = true;
                }
                catch (Exception exception)
                {
                    MessageBox.Show("Failed to start application. \n\nError: " + exception.Message, "Application Start Failed", MessageBoxButtons.OK,MessageBoxIcon.Asterisk);
                }
            }

            return success;
        }
        
        /// <summary>
        /// Sets up the image preview element, which couldn't be added through the designer.
        /// </summary>
        private void setUpPreview()
        {
            previewElementHost.Child = previewImage;
        }

        /// <summary>
        /// Loads an image into the preview image box.
        /// </summary>
        /// <param name="s">The image source of the image.</param>
        private void loadPreview(System.Windows.Media.ImageSource s)
        {
            previewImage.Source = s;
        }

        /// <summary>
        /// Updates the image file type from the current form state.
        /// </summary>
        private void updateOutputImageFileType()
        {
            OutputImageFileType = getImageFileType();
        }

        /// <summary>
        /// Updates the post process type from the current form state.
        /// </summary>
        private void updatePostProcess()
        {
            PostProcess = getPostProcessType();
        }

        /// <summary>
        /// Changes the extension of the output filename to match the type of file that is produced.
        /// </summary>
        private void updateOutputFilenameExtension()
        {
            OutputFilenameBase = System.IO.Path.ChangeExtension(OutputFilenameBase, getImageExt());
        }

        /// <summary>
        /// Updates the status display to show the user what is currently being processed.
        /// </summary>
        /// <param name="page">The current page.</param>
        /// <param name="pages">The number of pages in the entire document.</param>
        /// <param name="lastFilename">The filename of the image that was saved.</param>
        private void updateProgress(uint page, uint pages, string lastFilename = "")
        {
            double percent = 0;

            if (pages != 0)
            {
                percent = (double)page / (double)pages;
            }

            toolStripProgressBar.ProgressBar.Minimum = 0;
            toolStripProgressBar.ProgressBar.Maximum = 100;
            toolStripProgressBar.ProgressBar.Value = (int)(percent * 100);

            pageToolStripStatusLabel.Text = "Page " + page + " of " + pages + ".";

            filenameToolStripStatusLabel.Text = lastFilename;

            Application.DoEvents(); // Hack to make application look responsive, should really be in its own thread, but this is quick and dirty and does what I need.
        }

        /// <summary>
        /// Loads the persistant settings.
        /// </summary>
        private void loadSettings()
        {
            // Note: See Settings.settings file.
            // Note: Would have been nice to save enums raw, but had trouble with that so converted to string. It seems it should be possible but I couldn't work it out quickly.
            // Note: These settings could have been saved in their properties (rather then having local variables) but I just decided to do it this way.

            AlternateApplication = Properties.Settings.Default.alternateApplication;
            OutputImageFileType = (ImageType)Enum.Parse(typeof(ImageType), Properties.Settings.Default.outputImageFileType, true);
            OutputDPI = Properties.Settings.Default.outputDPI;
            PngInterlace = Properties.Settings.Default.pngInterlace;
            JpgQuality = Properties.Settings.Default.jpgQuality;
            UseInputFilename = Properties.Settings.Default.useInputFilename;
            CreateSubFolder = Properties.Settings.Default.createSubFolder;
            UseCurrentDateInFilename = Properties.Settings.Default.useCurrentDateInFilename;
            OutputFilenameBase = Properties.Settings.Default.outputFilenameBase;
            PostProcess = (PostProcessType)Enum.Parse(typeof(PostProcessType), Properties.Settings.Default.postProcess, true);
            PostProcessApplication = Properties.Settings.Default.postProcessApplication;
            DelayProcess = Properties.Settings.Default.delayProcess;
            DelaySeconds = Properties.Settings.Default.delaySeconds;
            DeleteWhenDone = Properties.Settings.Default.deleteWhenDone;
            AutoStart = Properties.Settings.Default.autoStart;
            AutoClose = Properties.Settings.Default.autoClose;

            fixRootOutputFilenameBase();
        }

        /// <summary>
        /// Saves the persistant settings.
        /// </summary>
        private void saveSettings()
        {
            // Note: See Settings.settings file.
            // Note: Would have been nice to save enums raw, but had trouble with that so converted to string. It seems it should be possible but I couldn't work it out quickly.
            // Note: These settings could have been saved in their properties (rather then having local variables) but I just decided to do it this way.

            Properties.Settings.Default.alternateApplication = AlternateApplication;
            Properties.Settings.Default.outputImageFileType = OutputImageFileType.ToString();
            Properties.Settings.Default.outputDPI = OutputDPI;
            Properties.Settings.Default.pngInterlace = PngInterlace;
            Properties.Settings.Default.jpgQuality = JpgQuality;
            Properties.Settings.Default.useInputFilename = UseInputFilename;
            Properties.Settings.Default.createSubFolder = CreateSubFolder;
            Properties.Settings.Default.useCurrentDateInFilename = UseCurrentDateInFilename;
            Properties.Settings.Default.outputFilenameBase = OutputFilenameBase;
            Properties.Settings.Default.postProcess = PostProcess.ToString();
            Properties.Settings.Default.postProcessApplication = PostProcessApplication;
            Properties.Settings.Default.delayProcess = DelayProcess;
            Properties.Settings.Default.delaySeconds = DelaySeconds;
            Properties.Settings.Default.deleteWhenDone = DeleteWhenDone;
            Properties.Settings.Default.autoStart = AutoStart;
            Properties.Settings.Default.autoClose = AutoClose;

            Properties.Settings.Default.Save();
        }

        /// <summary>
        /// Gets the image file state from the forms current status.
        /// </summary>
        /// <returns>The image file type currently selected on the form.</returns>
        private ImageType getImageFileType()
        {
            if (outputBMPRadioButton.Checked == true)
            {
                return ImageType.bmp;
            }
            if (outputPNGRadioButton.Checked == true)
            {
                return ImageType.png;
            }
            if (outputJPGRadioButton.Checked == true)
            {
                return ImageType.jpg;
            }

            return ImageType.bmp; // ERROR - Default To BMP
        }

        /// <summary>
        /// Gets the post processing state from the forms current status.
        /// </summary>
        /// <returns>The post processing type currently selected on the form.</returns>
        private PostProcessType getPostProcessType()
        {
            if (postProcessNothingRadioButton.Checked == true)
            {
                return PostProcessType.Nothing;
            }
            if (postProcessSingleRadioButton.Checked == true)
            {
                return PostProcessType.Single;
            }
            if (postProcessAllSequentialRadioButton.Checked == true)
            {
                return PostProcessType.AllSequential;
            }
            if (postProcessAllSimultaneousRadioButton.Checked == true)
            {
                return PostProcessType.AllSimultaneous;
            }

            return PostProcessType.Nothing; // ERROR - Default To Nothing
        }

        /// <summary>
        /// Gets the image extsionion for currently selected ouput image file type.
        /// 
        /// Note: Use the overload function to obtain the image file type for a specific file type.
        /// </summary>
        /// <returns>The file extension.</returns>
        private string getImageExt()
        {
            return getImageExt(OutputImageFileType);
        }

        //overloads ..type to get type assoicated with the type
        /// <summary>
        /// Gets the image extension for the image type specified.
        /// 
        /// Note: Use the overload function to obtain the image file type for the current file type.
        /// </summary>
        /// <param name="imageType">The image file type to obtain the extension of.</param>
        /// <returns>The extension of the image file type.</returns>
        private string getImageExt(ImageType imageType)
        {
            string ext = "";

            // Could replace this switch with ext = ". " imageType.ToString();
            switch (imageType)
            {
                case ImageType.bmp:
                    ext = bmpExt;
                    break;
                case ImageType.png:
                    ext = pngExt;
                    break;
                case ImageType.jpg:
                    ext = jpgExt;
                    break;
            }

            return ext;
        }

        /// <summary>
        /// Used to ensure that the filename for the output file is a fully rooted path.
        /// </summary>
        private void fixRootOutputFilenameBase()
        {
            // If we are loading an unrooted filename put it in the users pictures folder.
            if (!System.IO.Path.IsPathRooted(OutputFilenameBase) && (System.IO.Path.GetFileName(OutputFilenameBase) != ""))
            {
                OutputFilenameBase = System.IO.Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.MyPictures), OutputFilenameBase);
            }
        }


    }
}